angular.module('beamng.apps')

.directive('velocityForcedInduction', ['bngApi', 'StreamsManager', "UiUnits", function(bngApi, StreamsManager, UiUnits) {
  // noinspection CssUnknownTarget
  return {
    template:
        '<div>' +
          '<link rel="stylesheet" href="../ui/modules/apps/VelocityForcedInduction/app.css">' +
          '<link rel="stylesheet" href="../ui/tachometers/base.css">' +
          // Pre-load font
          '<link rel="preload" href="../ui/tachometers/fonts/Speed12.ttf" as="font" crossorigin="anonymous">' +
          '<canvas class="canvas" id="VFI_bgCanvas" width="150" height="150"></canvas>' +
          '<canvas class="canvas" id="VFI_canvas"   width="150" height="150"></canvas>' +
        '</div>',
    replace: true,
    restrict: 'EA',
    link: function ($scope, $element) {
      const ID = "VFI";
      const REQ_STREAMS = ["forcedInductionInfo"];
      const REQ_SERVICES = ["service", "velocityService"];

      StreamsManager.add(REQ_STREAMS);


      // Below has to be in all apps, and cannot be added to the BaseApp class since the file that
      // is in hasn't been loaded until this code has run

      let scriptsLoaded = 0;
      // Add in the script files for the base service
      for (let service of REQ_SERVICES) {
        let script = document.createElement("script");
        script.src = "../ui/tachometers/" + service + ".js";
        // Add in this directives ID to avoid conflicts between all of these modded apps
        script.id = service + ID;
        $element.append(script);

        document.getElementById(script.id).onload = function() {
          scriptsLoaded++;
          // Wait for all scripts to load
          if (scriptsLoaded === REQ_SERVICES.length)
            onload();
        };
      }


      /////////////
      // Imports //
      /////////////

      // Provisionally set to null - will be requested and set after element load
      let ModVelocityBase = null;
      let TachoBase = null;
      let app = null;

      ///////////////
      // Constants //
      ///////////////

      // The width of stroke effects for various components
      const WIDTH = {
        // The current boost arc that goes around the outside
        BOOST_ARC: 20,
        // The width of each tick
        TICK_WIDTH: 2
      }

      // The centre radius values for various components
      // Any stroke effect adds thickness to a radius
      const RADII = {
        BASE: 75,
        DIAL_RADIUS: 10
      };
      RADII.BOOST_ARC = RADII.BASE - WIDTH.BOOST_ARC/2;
      RADII.MIDDLE = 40;
      RADII.BOOST_NUMBERS = (RADII.BASE - WIDTH.BOOST_ARC + RADII.MIDDLE)/2;

      WIDTH.BOOST_NUMBERS = RADII.BASE - WIDTH.BOOST_ARC - RADII.MIDDLE;


      // The number of ticks between two numbers on the gauge
      const NUM_TICKS_PER_GAP = 3;
      // The number of numbers on the gauge to display
      const NUM_NUMBERS = 5;

      //////////////////////////////
      // Rendering/Main Functions //
      //////////////////////////////

      /**
       * Returns a list of all dynamic variables that will/may change between frames for the vehicle
       * @param {Object[]} forcedInductionInfo - the forced induction information stream
       */
      function getDynamicVars(forcedInductionInfo) {
        return {
          boost: forcedInductionInfo.boost
        };
      }


      $scope.$on("streamsUpdate", function (event, streams) {
        let forcedInductionInfo = streams.forcedInductionInfo;

        // Wait until ready
        if (!app || app.state === app.states.preInit || !forcedInductionInfo)
          return;

        let baseVars = getDynamicVars(forcedInductionInfo);

        // Perform first time setup if necessary
        if (app.state === app.states.init) {
          app.initFields(baseVars.boost, forcedInductionInfo.maxBoost);
          app.load(bngApi);
          app.drawBackgroundCanvas();
          app.state = app.states.loaded;
        }

        app.setCurrBoost(baseVars.boost);
        app.draw();
      });

      $scope.$on("$destroy", function () {
        StreamsManager.remove(REQ_STREAMS);
      });

      $scope.$on("VehicleFocusChanged", () => reload(true));
      $scope.$on("VehicleChange", () => reload(false));


      function reload(instantReload) {
        if (!app)
          return;

        let time = instantReload ? 0 : 50;
        setTimeout(() => {
          // It is unknown whether this vehicle has forced induction, so hide it provisionally and
          // it is does, the app will re-show itself
          app.hide();
          app.state = app.states.init;
        }, time);
      }


      /**
       * Loads the external services and initialises some base constants.
       */
      function onload() {
        // Request and setup the services
        let injector = angular.injector(["beamng.apps"]);
        ModVelocityBase = injector.get("ModVelocityBase");
        TachoBase = ModVelocityBase.TachoBase;


        /**
         * The concrete class representation of this app
         */
        class VelocityBoostApp extends TachoBase.BoostApp {
          constructor(boostAngleRange, baseRadius) {
            super(ID, boostAngleRange, baseRadius, document, bngApi);
          }

          /**
           * @Override
           * @inheritDoc
           */
          createGauge() {
            // Create concrete class of base class builder
            class TachoVelocityGaugeBuilder extends ModVelocityBase.VelocityGaugeBuilder {
              constructor(baseRadius, colourScheme, angleRange) {
                super(baseRadius, colourScheme, angleRange);
              }

              /**
               * @Override
               * Creates the GaugeArcs instance for the gauge
               * @param arcRadius
               * @param arcWidth
               * @param redlineAngle
               * @returns {GaugeBuilder} the GaugeBuilder instance
               */
              createArcs(arcRadius, arcWidth, redlineAngle) {
                return this.setArcs(new TachoBase.GaugeArcs(
                    new TachoBase.Arc(
                        arcRadius,
                        // Will be overridden later, so default to no arc
                        new TachoBase.Range(this.angleRange.lower, this.angleRange.lower),
                        new TachoBase.Style(this.colourScheme.ACCENT.secondary, arcWidth)
                    )
                ));
              }
            }


            let unitMaxBoost = UiUnits.pressure(this.maxBoost, this.unit).val;
            let floorMaxBoost = Math.floor(unitMaxBoost + 1);
            // Increment between numbers on the gauge
            let numberInc = floorMaxBoost / Math.floor(NUM_NUMBERS / 2);

            let numTicks = NUM_TICKS_PER_GAP*(NUM_NUMBERS-1) + NUM_NUMBERS;
            let tickSpacing = numberInc / (NUM_TICKS_PER_GAP + 1);

            this.gauge = new TachoVelocityGaugeBuilder(RADII.BASE, this.colourScheme, this.BOOST_ANGLE_RANGE)
                .createPointer(RADII.DIAL_RADIUS)
                // +1 to include 0 boost
                .createTicks(numTicks, tickSpacing, WIDTH.BOOST_ARC, -unitMaxBoost)
                .createArcs(RADII.BOOST_ARC, WIDTH.BOOST_ARC)
                .createNumbers(NUM_NUMBERS, 10, RADII.BOOST_NUMBERS, numberInc, -floorMaxBoost)
                .setCentre(this.CENTRE)
                .build();
          }

          /**
           * Draws/updates the background canvas
           */
          drawBackgroundCanvas() {
            super.drawBackgroundCanvas();
            const STROKE_WIDTH = 6;
            const RADIUS = RADII.MIDDLE - STROKE_WIDTH/2;

            let ctx = this.BG_CANVAS.get2DContext();

            // The inner circle part
            new TachoBase.Wedge(RADII.MIDDLE, this.BOOST_ANGLE_RANGE, TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE)
                .draw(ctx);
            // Draw interior arc between boost arc background and the interior
            new TachoBase.Arc(RADIUS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, STROKE_WIDTH), this.CENTRE)
                .draw(ctx);

            // The wedge not covered by the gauge area
            new TachoBase.Wedge(RADII.BASE, this.BOOST_ANGLE_RANGE.invert(), TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE)
                .draw(ctx);

            // The background arc to the RPM numbers
            new TachoBase.Arc(RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.BASE, WIDTH.BOOST_NUMBERS+1), this.CENTRE)
                .draw(ctx);
            // Draw background arc to boost gauge
            // Note: -1 to reduce overlap when the other canvas draws over it
            new TachoBase.Arc(RADII.BOOST_ARC, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, WIDTH.BOOST_ARC), this.CENTRE)
                .draw(ctx);

            // Draw the pointer dial as this only needs to be drawn once
            this.gauge.pointer.pointerDial.draw(ctx);

            this.drawUnit();
          }

          /**
           * Draws the boost's unit
           */
          drawUnit() {
            this.drawUnitBackground();
            this.drawUnitText();
          }

          /**
           * Draws the background of the unit area
           */
          drawUnitBackground() {
            let ctx = this.BG_CANVAS.get2DContext();
            ModVelocityBase.drawRoundRectRight(ctx, this, 90);
            ctx.fill();
          }

          /**
           * Draws the unit's name in text form
           */
          drawUnitText() {
            const FONT_SIZE = 20;
            const SPACING = 10;

            let ctx = this.BG_CANVAS.get2DContext();

            let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE.lower * 0.275);
            centrePos.addY(SPACING);

            new TachoBase.Text(new TachoBase.Font("Speed12", FONT_SIZE, true), this.unit, centrePos, this.colourScheme.TEXT.primary)
                .draw(ctx);
          }

          /**
           * @Override
           * @inheritDoc
           */
          draw() {
            super.draw();
            let ctx = this.CANVAS.get2DContext();

            let boost = UiUnits.pressure(this.currBoost, this.unit).val;

            // Normalise from range [-1,1] to [0,1]
            let normBoost = (this.currBoost/this.maxBoost + 1)/2;
            let currBoostAngle = this.BOOST_ANGLE_RANGE.getLinearRatio(normBoost);

            this.gauge.setAngle(currBoostAngle);
            this.gauge.draw(ctx, boost);
          }
        }

        app = new VelocityBoostApp(new TachoBase.Range(0.5*Math.PI, 2*Math.PI), RADII.BASE);
      }
    }
  };
}]);