angular.module('beamng.apps')
.directive('velocityTacho', ["StreamsManager", "UiUnits", "bngApi", function(StreamsManager, UiUnits, bngApi) {
  // noinspection CssUnknownTarget
  return {
    template:
        '<div>' +
          '<link rel="stylesheet" href="../ui/modules/apps/VelocityTacho/app.css">' +
          '<link rel="stylesheet" href="../ui/tachometers/base.css">' +
          // Pre-load font
          '<link rel="preload" href="../ui/tachometers/fonts/Speed12.ttf" as="font" crossorigin="anonymous">' +
          '<canvas class="canvas" id="VT_bgCanvas" width="300" height="300"></canvas>' +
          '<canvas class="canvas" id="VT_canvas"   width="300" height="300"></canvas>' +
          // '<div ng-include="\'../ui/tachometers/colours.html\'"></div>' +
        '</div>',
    replace: true,
    restrict: 'EA',
    link: function ($scope, $element) {
      // The internal ID for this app
      const ID = "VT";
      const REQ_STREAMS = ["electrics", "engineInfo"];
      const REQ_SERVICES = ["service", "velocityService"]

      StreamsManager.add(REQ_STREAMS);


      let scriptsLoaded = 0;
      // Add in the script files for the base service
      for (let service of REQ_SERVICES) {
        let script = document.createElement("script");
        script.src = "../ui/tachometers/" + service + ".js";
        // Add in this directives ID to avoid conflicts between all of these modded apps
        script.id = service + ID;
        $element.append(script);

        document.getElementById(script.id).onload = function() {
          scriptsLoaded++;
          // Wait for all scripts to load
          if (scriptsLoaded === REQ_SERVICES.length)
            onload();
        };
      }


      /////////////
      // Imports //
      /////////////

      // Provisionally set to null - will be requested and set after element load
      let ModVelocityBase = null;
      let TachoBase = null;
      let app = null;

      ///////////////
      // Constants //
      ///////////////

      // The width of stroke effects for various components
      const WIDTH = {
        // The current RPM arc that goes around the outside
        RPM_ARC: 30,
        // The interior's stroke between itself and the RPM background
        INNER_STROKE: 5,
        // The width of each tick on the RPM gauge
        RPM_TICK_WIDTH: 2
      }

      // The background to the RPM thousand numbers
      WIDTH.RPM_BG = 80 - WIDTH.RPM_ARC;

      /*
       * The centre radius values for various components.
       * It should be noted that the radii values are with respect to an arc that they represent.
       * So, a radius value is not the radius of that element, but the radius of the arc around the
       * centre of the UI it will be drawn as.
       * Any stroke effect adds thickness to a radius
       */
      const RADII = {
        // The radius of the whole UI
        BASE: 150,
        // The radius of the inner circle/background part
        BG_INNER: 90,
        // The radius of the pointer's dial
        RPM_DIAL_RADIUS: 30
      };
      // The radius of the stroke that separates the RPM numbers' background and the inner
      // background piece
      RADII.RPM_ARC_BG = RADII.BG_INNER + WIDTH.RPM_BG / 2;
      // The radius of the RPM arc that goes around the outside of the gauge
      RADII.RPM_ARC = RADII.BASE - WIDTH.RPM_ARC / 2;
      // The radius of the inner (stroked) part of the RPM arc
      RADII.RPM_INNER = RADII.BASE - WIDTH.RPM_ARC;
      // The radius of the inner background piece
      RADII.MIDDLE = RADII.BG_INNER - WIDTH.INNER_STROKE * 2;
      // The radius the RPM numbers are displayed around
      RADII.RPM_NUMBERS = RADII.RPM_INNER - (RADII.RPM_INNER - RADII.BG_INNER + WIDTH.INNER_STROKE) / 2;

      // The RPM spacing between two ticks
      const TICK_SPACING = 250;

      //////////////////////////////
      // Rendering/Main Functions //
      //////////////////////////////

      function getMaxEngineRPM(engineInfo) {
        return engineInfo[1];
      }

      /**
       * Returns a list of all dynamic variables that will/may change between frames for the vehicle
       * @param {Object} engineInfo - the engine info object from the engineInfo stream
       * @param {Object} electrics - the electrics object from the electrics stream
       * @returns {{currSpeed: number, currRPM: number, lights: number, currGear: number,
       * pBrakeEngaged: boolean}}
       */
      function getDynamicVars(engineInfo, electrics) {
        let lights = 0;
        if (electrics.highbeam)
          lights = 2;
        else if (electrics.lowbeam)
          lights = 1;

        return {
          currGear: engineInfo[16],
          currSpeed: electrics.wheelspeed,
          currRPM: engineInfo[4],
          lights: lights,
          pBrakeEngaged: electrics.parkingbrake > 0
        };
      }


      $scope.$on("streamsUpdate", function (event, streams) {
        let engineInfo = streams.engineInfo;
        let electrics = streams.electrics;

        // Ensure engineInfo exists AND ALSO that the engine is actually setup at this point
        let isEngineSetup = engineInfo && getMaxEngineRPM(engineInfo) !== 0;

        // Wait until everything is ready
        if (!app || app.state === app.states.preInit || !isEngineSetup || !electrics)
          return;

        let baseVars = getDynamicVars(engineInfo, electrics);

        // Perform first time setup if necessary
        if (app.state === app.states.init) {
          app.initFields(baseVars.currGear, baseVars.currSpeed, baseVars.currRPM, getMaxEngineRPM(engineInfo), baseVars.lights);
          app.load(bngApi);
          app.drawBackgroundCanvas();
          app.state = app.states.loaded;
        }

        app.updateVariables(baseVars.currGear, baseVars.currSpeed, baseVars.currRPM, baseVars.lights, baseVars.pBrakeEngaged);
        app.draw();
      });

      $scope.$on("$destroy", function () {
        StreamsManager.remove(REQ_STREAMS);
      });

      $scope.$on("VehicleFocusChanged", () => reload(true));
      $scope.$on("VehicleChange", () => reload(false));


      function reload(instantReload) {
        if (!app)
          return;

        let time = instantReload ? 0 : 50;
        setTimeout(() => {app.state = app.states.init;}, time);
      }

      /**
       * Loads the external services and initialises some base constants.
       */
      function onload() {
        // Request and setup the services
        let injector = angular.injector(["beamng.apps"]);
        ModVelocityBase = injector.get("ModVelocityBase");
        TachoBase = ModVelocityBase.TachoBase;


        /**
         * The concrete class representation of this app
         */
        class VelocityTachoApp extends TachoBase.TachometerApp {
          constructor(rpmAngleRange, baseRadius) {
            super(ID, rpmAngleRange, baseRadius, document, bngApi);
            this.needsRedraw = false;
          }

          /**
           * @Override
           * @inheritDoc
           */
          updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged) {
            let prevState = this.collateBGVars();
            super.updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged);
            let newState = this.collateBGVars();

            // If any variables for the background canvas have changed, the background canvas
            // requires a re-draw
            if (!prevState.every(function(v, i) { return v === newState[i]}))
              this.needsRedraw = true;
          }

          /**
           * @returns {[*]} the list of all variables that drawn on the background canvas. This
           * includes the gear index, the lights' state and the parking brake
           */
          collateBGVars() {
            return [this.currGearIndex, this.lightsState, this.pBrakeEngaged];
          }

          /**
           * @Override
           * @inheritDoc
           */
          createGauge() {
            // Create concrete class of base class builder
            class TachoVelocityGaugeBuilder extends ModVelocityBase.VelocityGaugeBuilder {
              constructor(baseRadius, colourScheme, angleRange) {
                super(baseRadius, colourScheme, angleRange);
              }

              /**
               * @Override
               * Creates the GaugeArcs instance for the gauge
               * @param arcRadius
               * @param arcWidth
               * @param redlineAngle
               * @returns {GaugeBuilder} the GaugeBuilder instance
               */
              createArcs(arcRadius, arcWidth, redlineAngle) {
                return this.setArcs(new TachoBase.GaugeArcs(
                    new TachoBase.Arc(
                        arcRadius,
                        // Will be overridden later, so default to no arc
                        new TachoBase.Range(this.angleRange.lower, this.angleRange.lower),
                        new TachoBase.Style(this.colourScheme.ACCENT.secondary, arcWidth)
                    ),
                    new TachoBase.Arc(
                        arcRadius,
                        new TachoBase.Range(redlineAngle, this.angleRange.upper),
                        new TachoBase.Style("red", arcWidth)
                    )
                ));
              }
            }

            this.gauge = new TachoVelocityGaugeBuilder(this.radius, this.colourScheme, this.RPM_ANGLE_RANGE)
                .createPointer(RADII.RPM_DIAL_RADIUS)
                // +1 to include 0RPM
                .createTicks(Math.floor(this.redLineRPM / TICK_SPACING) + 1, TICK_SPACING, WIDTH.RPM_ARC)
                .createArcs(RADII.RPM_ARC, WIDTH.RPM_ARC, this.getRedlineRPMAngle())
                // +1 to include 0
                .createNumbers(this.redLineRPM/1000 + 1, 20, RADII.RPM_NUMBERS, 1)
                .setCentre(this.CENTRE)
                .build();
          }

          /**
           * @Override
           * @inheritDoc
           * This includes both backgrounds for the speedometer as well as the divider between the
           * RPM gauge and inner part. This also draws the background parts of the RPM (see
           * setupRPM()) and draws the gear in the centre (see drawGear())
           */
          drawBackgroundCanvas() {
            super.drawBackgroundCanvas();
            let ctx = this.BG_CANVAS.get2DContext();

            // The background wedge for the last quarter not filled by the RPM
            new TachoBase.Wedge(RADII.BASE, this.RPM_ANGLE_RANGE.invert(), TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE)
                .draw(ctx);
            // The inner circle part
            new TachoBase.Wedge(RADII.MIDDLE, this.RPM_ANGLE_RANGE, TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE)
                .draw(ctx);

            this.setupRPM();
            this.drawGear();
            this.drawSpeedBackground();
            this.drawParkingBrake();
          }

          /**
           * Setups and draws the RPM gauge's background canvas objects.
           * This draws the dividing arcs as well as the backgrounds (such as redline RPM) for the RPM.
           * This also draws the inner parts of the RPM pointer as well as generating the RPM ticks
           */
          setupRPM() {
            const STROKE_WIDTH = 10;
            const RADIUS = RADII.BG_INNER - STROKE_WIDTH / 2;

            let ctx = this.BG_CANVAS.get2DContext();
            let redlineRPMAngle = app.getRedlineRPMAngle();

            // Draw interior arc between RPM background and the interior
            new TachoBase.Arc(RADIUS, app.RPM_ANGLE_RANGE, new TachoBase.Style(app.colourScheme.ACCENT.primary, STROKE_WIDTH), app.CENTRE)
                .draw(ctx);
            // The background arc to the RPM numbers
            new TachoBase.Arc(RADII.RPM_ARC_BG, app.RPM_ANGLE_RANGE, new TachoBase.Style(app.colourScheme.BASE, WIDTH.RPM_BG + 1), app.CENTRE)
                .draw(ctx);

            // Draw non-redline background RPM
            new TachoBase.Arc(RADII.RPM_ARC, new TachoBase.Range(app.RPM_ANGLE_RANGE.lower, redlineRPMAngle), new TachoBase.Style(app.colourScheme.ACCENT.primary, WIDTH.RPM_ARC), app.CENTRE)
                .draw(ctx);

            // Draw the pointer dial as this only needs to be drawn once
            this.gauge.pointer.pointerDial.draw(ctx);
          }

          /**
           * Draws the gear number in the middle of the tachometer
           */
          drawGear() {
            let gearName = TachoBase.TachometerApp.gearIndexToName(this.currGearIndex);
            // Base font size is 40; if there is more >1 digit on the gear then decrease to fit
            let fontSize = 40 / (1 + (gearName.length - 1) * 0.5);

            let ctx = this.BG_CANVAS.get2DContext();

            VelocityTachoApp.drawPrimaryText(ctx, gearName, fontSize, this.CENTRE);
            VelocityTachoApp.resetFilter(ctx);
          }

          /**
           * Draws the background polygon for the speedometer
           */
          drawSpeedBackground() {
            let ctx = this.BG_CANVAS.get2DContext();
            ModVelocityBase.drawRoundRectRight(ctx, this, 170);

            // Apply lighting glow
            if (this.lightsState > 0) {
              // Green for low beam, blue for high beam
              let colour = this.lightsState === 1 ? "green" : "rgb(0, 100, 255)";
              ctx.filter = "drop-shadow(-8px -8px 8px " + colour + ")";
            }

            ctx.fill();
            VelocityTachoApp.resetFilter(ctx);
          }

          getPBrakePolygonAngles() {
            return new TachoBase.Range(
                this.RPM_ANGLE_RANGE.lower * 0.475,
                this.RPM_ANGLE_RANGE.lower * 0.825
            );
          }

          /**
           * Draws the background polygon of the parking brake text
           */
          drawParkingBrake() {
            let ctx = this.BG_CANVAS.get2DContext();
            ctx.beginPath();

            new TachoBase.RoundCornerRectangleBR(this.radius, this.getPBrakePolygonAngles(), this.CENTRE, 170)
                .drawContinue(ctx);

            ctx.lineWidth = 1;
            ctx.fillStyle = "black";
            ctx.fill();

            this.drawParkingBrakeText();
          }

          /**
           * Draws the parking brake text, coloured to indicate whether it is applied or not
           */
          drawParkingBrakeText() {
            let colour = this.pBrakeEngaged ? "orange" : "rgb(24, 24, 24)";

            let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, RADII.RPM_ARC, this.getPBrakePolygonAngles().getLinearRatio(0.5));
            // Manual tuning to make it look more central
            centrePos.addY(8);
            centrePos.addX(-20);

            let ctx = this.BG_CANVAS.get2DContext();

            // Apply glow
            if (this.pBrakeEngaged)
              ctx.filter = "drop-shadow(0 0 8px " + colour + ")";

            new TachoBase.Text(new TachoBase.Font("Speed12", 15, true), "PARK", centrePos, colour)
                .draw(ctx);
            VelocityTachoApp.resetFilter(ctx);
          }

          /**
           * @Override
           * @inheritDoc
           * Draws the RPM gauge and the interior
           */
          draw() {
            super.draw();
            this.drawRPM();
            this.drawInterior();
          }

          /**
           * Draws the interior parts of the tachometer. This includes the speedometer
           */
          drawInterior() {
            // Re-draw background canvas if a redraw is required
            if (this.needsRedraw) {
              this.drawBackgroundCanvas();
              this.needsRedraw = false;
            }

            let convertedSpeed = UiUnits.speed(this.speed, this.unit);
            this.drawSpeed(Math.floor(convertedSpeed.val), convertedSpeed.unit);
          }

          /**
           * Draws the speedometer and its units
           * @param {number} speed - the speed of the vehicle
           * @param {string} unit - the unit {@param speed} is in
           */
          drawSpeed(speed, unit) {
            // The spacing between pieces of text and the space between the top of the speed-box and
            // the first piece of text
            const SPACING = 5;
            const SPEED_FONT_SIZE = 40;
            const UNIT_FONT_SIZE = 20;
            let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, RADII.BG_INNER, ModVelocityBase.roundRectRightAngle(this).getLinearRatio(0.5));

            let ctx = this.CANVAS.get2DContext();

            // Centre the text
            centrePos.addY(SPACING * 3);
            VelocityTachoApp.drawPrimaryText(ctx, speed.toString(), SPEED_FONT_SIZE, centrePos);
            // Account for the size of the speed text to ensure no-overlap
            centrePos.addY(SPEED_FONT_SIZE / 2 + SPACING);
            VelocityTachoApp.drawPrimaryText(ctx, unit, UNIT_FONT_SIZE, centrePos);
          }

          /**
           * Draws the RPM gauge. This includes the RPM ticks, pointer and current RPM arc
           */
          drawRPM() {
            let currRPMAngle = this.RPM_ANGLE_RANGE.getLinearRatio(this.currRPM / this.redLineRPM);
            this.gauge.setAngle(currRPMAngle);
            this.gauge.draw(this.CANVAS.get2DContext(), this.currRPM);
          }


          /**
           * Resets the filter to prevent lingering filters
           */
          static resetFilter(ctx) {
            ctx.filter = "none";
          }

          /**
           * Draws centred text in the colour scheme's primary text colour at the position specified
           * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
           * @param {string} text - the text to write
           * @param {number} fontSize - the font size in px
           * @param {Vector} position - the position to set the centre of the text to
           */
          static drawPrimaryText(ctx, text, fontSize, position) {
            new TachoBase.Text(
                new TachoBase.Font("Speed12", fontSize, true),
                text,
                position,
                app.colourScheme.TEXT.primary
            ).draw(ctx);
          }
        }

        app = new VelocityTachoApp(new TachoBase.Range(0.5*Math.PI, 2*Math.PI), RADII.BASE);
      }
    }
  };
}]);