/////////////
// Classes //
/////////////

/**
 * A class representing a 2D vector
 */
class Vector {
  constructor(x, y) {
    this.array = [x, y];
  }

  /**
   * Copies a vector and returns its copy
   * @param {Vector} v - the vector to copy
   * @returns {Vector} the copied vector instance
   */
  static copy(v) {
    return new Vector(v.array[0], v.array[1]);
  }

  /**
   * Reverse the signs of all elements in the vector
   */
  negate() {
    this.array.forEach(e => -e);
  }

  static mapOperations(vector, amount, func) {
    let result = new Vector(vector.array[0], vector.array[1]);

    if (amount instanceof Vector) {
      for (let i = 0; i < result.array.length; i++)
        result.array[i] = func(result.array[i], amount.array[i]);
    } else {
      result.array = result.array.map((e) => func(e, amount));
    }

    return result;
  }

  /**
   * Adds a vector from another or a scalar value
   * @param {Vector} vector - the vector to be added to
   * @param {Vector|number} amount - the vector/scalar to add
   * @returns {Vector} - the resulting vector
   */
  static add(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e+a);
  }

  /**
   * Subtracts a vector from another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static sub(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e-a);
  }

  /**
   * Multiples a vector with another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static mult(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e*a);
  }

  /**
   * Divides a vector by another or a scalar value
   * @param {Vector} vector - the vector to be divided by
   * @param {Vector|number} amount - the vector/scalar to divide with
   * @returns {Vector} - the resulting vector
   */
  static div(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e/a);
  }

  /**
   * Adds to this vector another vector from another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to add
   */
  add(amount) {
    this.array = Vector.add(this, amount).array;
  }

  /**
   * Subtracts this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  sub(amount) {
    this.array = Vector.sub(this, amount).array;
  }

  /**
   * Multiples this vector with another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  mult(amount) {
    this.array = Vector.mult(this, amount).array;
  }

  /**
   * Divides this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to divide with
   */
  div(amount) {
    this.array = Vector.div(this, amount).array;
  }

  /**
   * Normalises this vector
   */
  normalise() {
    let mag = this.mag();

    if (mag > 0)
      this.div(mag);
  }

  /**
   * Calculates and returns this vector's magnitude
   * @returns {number} - the magnitude of this vector
   */
  mag() {
    return Math.sqrt(this.array.reduce((a, b) => Math.pow(a, 2) + Math.pow(b, 2)));
  }

  /**
   * Sets this vector's components to that of another's
   * @param {Vector} vector - the components to copy
   */
  set(vector) {
    for (let i = 0; i < this.array.length; i++)
      this.array[i] = vector.array[i];
  }

  setX(x) { this.array[0] = x;}
  setY(y) { this.array[1] = y;}

  x() { return this.array[0]; }
  y() { return this.array[1]; }
}

///////////////
// Constants //
///////////////

// The width of stroke effects for various components
const WIDTH = {
  // The current RPM arc that goes around the outside
  RPM_ARC: 30,
  // The interior's stroke between itself and the RPM background
  INNER_STROKE: 5,
  // The width of each tick on the RPM gauge
  RPM_TICK_WIDTH: 2
}

// The background to the RPM thousand numbers
WIDTH.RPM_BG = 80 - WIDTH.RPM_ARC;

// The centre radius values for various components
// Any stroke effect adds thickness to a radius
const RADII = {
  BASE: 150,
  BG_INNER: 90,
  RPM_DIAL_RADIUS: 30
};
RADII.RPM_ARC_BG = RADII.BG_INNER + WIDTH.RPM_BG/2;
RADII.RPM_ARC = RADII.BASE - WIDTH.RPM_ARC/2;
RADII.RPM_INNER = RADII.BASE - WIDTH.RPM_ARC;


const CENTRE = new Vector(150, 150);

// The colour scheme
const ACCENT_COLOUR  = "rgb(255, 255, 255)";
const ACCENT_COLOUR2 = "rgb(  0,   0,   0)";
// const RPM_COLOUR     = "rgb(  0, 150, 255)";
const TEXT_COLOUR    = "rgb(255, 255,   0)";
const RPM_COLOUR     = TEXT_COLOUR;
const BASE_COLOUR    = "rgb( 64,  64,  64)";
const INNER_COLOUR   = "rgb( 96,  96,  96)";

// The starting and end angle of the RPM counter
const RPM_START_ANGLE = 0.5*Math.PI;
const RPM_END_ANGLE = 2*Math.PI;

/////////////////////
// Scope Variables //
/////////////////////

$scope = {};
$scope.needsSetup = false;
$scope.lightsState = 0;
$scope.currGearIndex = 0;
$scope.speed = 0;
$scope.currRPM = 0;
$scope.maxRPM = 0;
$scope.redLineRPM = 0;

//////////////////////////////
// Rendering/Main Functions //
//////////////////////////////

function setupCanvas(canvas) {
  let canvasElem = document.getElementById(canvas);
  // Ensure correct HD resolution on the canvas (fix for blurry text)
  canvasElem.width = canvasElem.getBoundingClientRect().width;
  canvasElem.height = canvasElem.getBoundingClientRect().height;
}

/**
 * Draws the static elements and setups the scope variables
 * @param {number} currGearIndex - the current gear's index the vehicle is in
 * @param {number} speed - the current speed of the car
 * @param {number} currRPM - the current RPM of the engine
 * @param {number} maxRPM - the maximum RPM of the engine
 * @param {number} lightsState - the state of the lights (0=off, 1=low, 2=high)
 */
function setup(currGearIndex, speed, currRPM, maxRPM, lightsState) {
  setupScopeVars(currGearIndex, speed, currRPM, maxRPM, lightsState);

  setupCanvas("bgCanvas");
  setupCanvas("canvas");

  drawBackgroundCanvas();
}

function drawBackgroundCanvas() {
  let ctx = document.getElementById("bgCanvas").getContext("2d");

  // The inner circle wedge which stores the misc icons
  drawWedge(ctx, CENTRE, RADII.BG_INNER - 10, RPM_START_ANGLE, RPM_END_ANGLE, INNER_COLOUR, 0, true);
  // The background wedge for the last quarter not filled by the RPM
  drawWedge(ctx, CENTRE, RADII.BASE, RPM_END_ANGLE, RPM_START_ANGLE, BASE_COLOUR, 0, true);

  setupRPM(ctx, $scope.maxRPM, $scope.redLineRPM, $scope.lightsState);
  drawGear(ctx, $scope.currGearIndex);
}

/**
 * Sets the scope variables
 * @param {number} currGearIndex - the current gear's index the vehicle is in
 * @param {number} speed - the current speed of the car
 * @param {number} currRPM - the current RPM of the engine
 * @param {number} maxRPM - the maximum RPM of the engine
 * @param {number} lightsState - the state of the lights (0=off, 1=low, 2=high)
 */
function setupScopeVars(currGearIndex, speed, currRPM, maxRPM, lightsState) {
  $scope.currGearIndex = currGearIndex;
  $scope.speed = speed;
  $scope.currRPM = currRPM;
  $scope.maxRPM = maxRPM;
  // The next rounded thousand of the max RPM
  $scope.redLineRPM = Math.floor($scope.maxRPM/1000 + 1) * 1000;
  $scope.lightsState = lightsState;
}

function setupRPM(ctx, maxRPM, redLineRPM, lightsState) {
  const STROKE_WIDTH = 10;
  const RADIUS = RADII.BG_INNER - STROKE_WIDTH/2;

  let redlineRPMAngle = getAngleOfRatio(RPM_START_ANGLE, RPM_END_ANGLE, maxRPM / redLineRPM);

  // Draw interior arc between RPM background and the interior
  drawArc(ctx, CENTRE, RADIUS, RPM_START_ANGLE, RPM_END_ANGLE, ACCENT_COLOUR2, STROKE_WIDTH, false);
  // The background arc showing the RPM areas
  drawArc(ctx, CENTRE, RADII.RPM_ARC_BG, RPM_START_ANGLE, RPM_END_ANGLE, BASE_COLOUR, WIDTH.RPM_BG);

  // Draw non-redline background RPM
  // Note: -1 to reduce overlap when the other canvas draws over it
  drawArc(ctx, CENTRE, RADII.RPM_ARC - 1, RPM_START_ANGLE, redlineRPMAngle, ACCENT_COLOUR2, WIDTH.RPM_ARC);
  // Draw the red-line RPM
  drawArc(ctx, CENTRE, RADII.RPM_ARC, redlineRPMAngle, RPM_END_ANGLE, "red", WIDTH.RPM_ARC);

  // Draw centre part of pointer
  setupRPMPointer(ctx);
  // The RPM ticks
  drawRPMTicks(ctx, redLineRPM, redlineRPMAngle);

  // Draw the current speed's background
  drawSpeedBackground(ctx, lightsState);
}

/**
 * Draws the centre part of the RPM pointer
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 */
function setupRPMPointer(ctx) {
  const STROKE_WIDTH = 5;
  // Draw centre circle part
  drawCircle(ctx, CENTRE, RADII.RPM_DIAL_RADIUS, ACCENT_COLOUR2, STROKE_WIDTH, true);
  // drawCircle(ctx, CENTRE, RADII.RPM_DIAL_RADIUS - STROKE_WIDTH/2, ACCENT_COLOUR2, 1, true);
}


function draw(currGear, currSpeed, currRPM, lightsState) {
  let canvas = document.getElementById("canvas");
  let ctx = canvas.getContext("2d");

  // Clear the last frame
  ctx.clearRect(0, 0, canvas.width, canvas.height);

  drawRPM(ctx, currRPM, $scope.redLineRPM, $scope.maxRPM);
  drawInterior(ctx, currGear, currSpeed, lightsState);
}

/**
 * Resets the filter to prevent lingering colouring
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 */
function resetLights(ctx) {
  ctx.filter = "none";
}

function drawInterior(ctx, currGearIndex, speed, lightsState) {
  let needsRedraw = false;

  // Update if changes have occurred
  if ($scope.currGearIndex !== currGearIndex) {
    $scope.currGearIndex = currGearIndex;
    needsRedraw = true;
  }
  if ($scope.lightsState !== lightsState) {
    $scope.lightsState = lightsState;
    needsRedraw = true;
  }

  if (needsRedraw)
    drawBackgroundCanvas();

  drawSpeed(ctx, speed);
}

function drawGear(ctx, currGearIndex) {
  let gearName;
  let fontSize = 40;

  if (currGearIndex === 0) {
    gearName = "N";
  } else if (currGearIndex < 0) {
    // Add index of reverse gear if more than one exists on the vehicle
    if (currGearIndex < -1) {
      gearName = "R" + Math.abs(currGearIndex);
      // Decrease font size for multiple characters
      fontSize /= 2;
    } else {
      gearName = "R";
    }
  } else {
    gearName = currGearIndex;
  }

  drawText(ctx, gearName, Vector.copy(CENTRE), fontSize, TEXT_COLOUR, true);
}

function drawSpeed(ctx, speed) {
  const SPACING = 10;
  const CENTRE_POS = getCircularPointRelCentre(RADII.BG_INNER, RPM_START_ANGLE/3);

  // Draw the speed text
  let speedPos = Vector.copy(CENTRE_POS);
  speedPos.setY(CENTRE_POS.y() + SPACING);
  drawText(ctx, speed, speedPos, 50, TEXT_COLOUR, true);

  // Draw the unit text
  let unitPos = Vector.copy(speedPos);
  unitPos.setY(speedPos.y() + SPACING);
  drawText(ctx, "MPH", unitPos, 10, TEXT_COLOUR, true);

  $scope.speed = speed;
}

function drawSpeedBackground(ctx, lightsState) {
  // The starting and end angles for the arc
  const START_ANGLE = RPM_START_ANGLE * 0.125;
  const END_ANGLE = RPM_START_ANGLE * 0.425;

  // The minimum value of x, controlling the inner-most part of the background
  const MIN_X = 170;

  ctx.beginPath();
  ctx.arc(CENTRE.x(), CENTRE.y(), RADII.BASE, START_ANGLE, END_ANGLE);
  ctx.lineTo(MIN_X, getCircularPointRelCentre(RADII.BASE, END_ANGLE).y());
  ctx.lineTo(MIN_X, getCircularPointRelCentre(RADII.BASE, START_ANGLE).y());

  ctx.lineWidth = "0";
  ctx.fillStyle = "black";

  // Apply lighting glow
  if (lightsState > 0) {
    // Green for low beam, blue for high beam
    let colour = lightsState === 1 ? "green" : "rgb(0, 100, 255)";
    ctx.filter = "drop-shadow(-8px -8px 8px " + colour + ")";
  }

  ctx.fill();
  resetLights(ctx);
}

function drawRPM(ctx, currRPM, redLineRPM, maxRPM) {
  // The angle the RPM value reaches
  let currRPMAngle = getAngleOfRatio(RPM_START_ANGLE, RPM_END_ANGLE, currRPM / redLineRPM);
  let redlineRPMAngle = getAngleOfRatio(RPM_START_ANGLE, RPM_END_ANGLE, maxRPM / redLineRPM);

  // The current RPM
  drawRPMStroke(ctx, currRPMAngle, redlineRPMAngle);
  // The pointer in the centre
  drawRPMPointer(ctx, currRPMAngle);
}

function getAngleOfRatio(startAngle, endAngle, ratio) {
  let dif = endAngle - startAngle;
  return startAngle + ratio*dif;
}

function drawRPMPointer(ctx, angleRPM) {
  // The width of the pointer at its maximum point
  const ANGLE_WIDTH = 0.1*Math.PI;
  // The radius of the smaller pointer on the opposite side
  const INV_POINTER_RADIUS = RADII.RPM_DIAL_RADIUS*1.5;

  // The angle at which the current RPM equals on the opposite side
  let angleRPMInv = angleRPM + Math.PI;

  // Draw triangular piece opposite to the RPM pointed at
  drawTriangleOnCircularBase(ctx, angleRPMInv, ANGLE_WIDTH, RADII.RPM_DIAL_RADIUS, INV_POINTER_RADIUS, ACCENT_COLOUR2, 1, true);
  // Draw triangular piece that points to the current RPM value
  drawTriangleOnCircularBase(ctx, angleRPM, ANGLE_WIDTH, RADII.RPM_DIAL_RADIUS, RADII.BASE, ACCENT_COLOUR2, 1, true);

  // Draw accents onto the pointer
  const ANGLE_WIDTH_ACCENTS = ANGLE_WIDTH/4;
  drawTriangleOnCircularBase(ctx, angleRPM, ANGLE_WIDTH_ACCENTS, RADII.RPM_DIAL_RADIUS, RADII.BASE, RPM_COLOUR, 1, true);
  drawTriangleOnCircularBase(ctx, angleRPMInv, ANGLE_WIDTH_ACCENTS, RADII.RPM_DIAL_RADIUS, INV_POINTER_RADIUS, RPM_COLOUR, 1, true);
}

function drawRPMStroke(ctx, angleRPM, redlineRPMAngle) {
  // Ensure it doesn't go past redline
  angleRPM = Math.min(angleRPM, redlineRPMAngle);
  // The current RPM
  drawArc(ctx, CENTRE, RADII.RPM_ARC, RPM_START_ANGLE, angleRPM, RPM_COLOUR, WIDTH.RPM_ARC);
}

function drawRPMTicks(ctx, maxRPM, redlineRPMAngle) {
  // The RPM spacing between two ticks
  const TICK_SPACING = 250;
  const NUM_TICKS = Math.floor(maxRPM / TICK_SPACING);
  // The angle spacing between two ticks
  const SPACING = (RPM_END_ANGLE - RPM_START_ANGLE) / NUM_TICKS;

  let angle = RPM_START_ANGLE;

  for (let i = 0; i <= NUM_TICKS; i++) {
    let rpmOfCurrAngle = i * TICK_SPACING;
    let isAThousand = rpmOfCurrAngle % 1000 === 0;

    // Label the thousand
    if (isAThousand) {
      let numThousand = rpmOfCurrAngle / 1000;
      labelRPMThousand(ctx, numThousand, angle, angle >= redlineRPMAngle);
    }

    // Add in the tick
    addRPMTick(ctx, angle, isAThousand, i === 0 || rpmOfCurrAngle === maxRPM);

    angle += SPACING;
  }
}

function getRadiusOfRPMAt(isInner, isPadded) {
  const PADDING = WIDTH.RPM_TICK_WIDTH/2;

  if (isInner) {
    let radius = RADII.RPM_INNER;
    return isPadded ? radius+PADDING : radius;
  } else {
    let radius = RADII.BASE;
    return isPadded ? radius-PADDING : radius;
  }
}

function labelRPMThousand(ctx, num, angle, isAtRedline) {
  const FONT_SIZE = 20;

  ctx.font = FONT_SIZE+"px Advantage";
  ctx.textAlign = "center";
  ctx.fillStyle = isAtRedline ? "red" : ACCENT_COLOUR;
  const FONT_OFFSET = new Vector(0, ctx.measureText("M").width/2);

  let coords = getCircularPointRelCentre(RADII.RPM_ARC_BG - FONT_SIZE/2, angle);
  coords.add(FONT_OFFSET);
  ctx.fillText(num, coords.x(), coords.y());
}

function addRPMTick(ctx, angle, isAThousand, isOuterPadded) {
  let r1 = getRadiusOfRPMAt(false, !isOuterPadded);
  let r2 = getRadiusOfRPMAt(true, false);
  // Half size for non-thousands
  if (!isAThousand)
    r2 += (r1-r2)/2;

  let p1 = getCircularPointRelCentre(r1, angle);
  let p2 = getCircularPointRelCentre(r2, angle);
  drawLine(ctx, p1, p2, ACCENT_COLOUR, WIDTH.RPM_TICK_WIDTH);
}

///////////////////////
// Drawing Functions //
///////////////////////

/**
 * Draws a line that connects two points
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} point1 - the first point
 * @param {Vector} point2 - the second point
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function drawLine(ctx, point1, point2, strokeColour, lineWidth) {
  ctx.beginPath();
  ctx.moveTo(point1.x(), point1.y());
  ctx.lineTo(point2.x(), point2.y());
  applyStyleAndStroke(ctx, strokeColour, lineWidth, false);
}

/**
 * Draws an arc between two angles
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the arc
 * @param {number} radius - the radius of the arc
 * @param {number} startAngle - the starting angle of the arc
 * @param {number} endAngle -  the ending angle of the arc
 * @param {string} strokeColour - the colour of the arc
 * @param {number} lineWidth - the width of the arc
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawArc(ctx, centre, radius, startAngle, endAngle, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.arc(centre.x(), centre.y(), radius, startAngle, endAngle);
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a circle at a specific point
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the circle
 * @param {number} radius - the radius of the circle
 * @param {string} strokeColour - the colour of the circle
 * @param {number} lineWidth - the thickness of the line of the circle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawCircle(ctx, centre, radius, strokeColour, lineWidth, filled) {
  drawArc(ctx, centre, radius, 0, 2*Math.PI, strokeColour, lineWidth, filled);
}

/**
 * Draws an arc between two angles and fills it in, relative to its centre origin, to create a
 * wedge shape
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the wedge
 * @param {number} radius - the radius of the wedge
 * @param {number} startAngle - the starting angle of the wedge
 * @param {number} endAngle -  the ending angle of the wedge
 * @param {string} strokeColour - the colour of the wedge
 * @param {number} lineWidth - the width of the wedge
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawWedge(ctx, centre, radius, startAngle, endAngle, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.moveTo(CENTRE.x(), CENTRE.y());
  ctx.arc(centre.x(), centre.y(), radius, startAngle, endAngle);
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a triangle, specified by its three vertices
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} point1 - the first vertex
 * @param {Vector} point2 - the second vertex
 * @param {Vector} point3- the second vertex
 * @param {string} strokeColour - the colour of the triangle
 * @param {number} lineWidth - the thickness of the line of the triangle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawTriangle(ctx, point1, point2, point3, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.moveTo(point1.x(), point1.y());
  ctx.lineTo(point2.x(), point2.y());
  ctx.lineTo(point3.x(), point3.y());
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a triangle whose base two vertices are positioned on a circle, and whose third vertex is
 * extended on another circle
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {number} centreAngle - the angle of the circle at the centre point of the triangle
 * @param {number} angleDev - the amount past the centre angle the triangle will extend to
 * @param {number} circleRadius - the circle's radius
 * @param {number} middleRadius - the radius the top-most vertex of the triangle will be placed on
 * @param {string} strokeColour - the colour of the triangle
 * @param {number} lineWidth - the thickness of the line of the triangle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawTriangleOnCircularBase(ctx, centreAngle, angleDev, circleRadius, middleRadius, strokeColour, lineWidth, filled) {
  let p1 = getCircularPointRelCentre(circleRadius, centreAngle - angleDev);
  let p2 = getCircularPointRelCentre(circleRadius, centreAngle + angleDev);
  let p3 = getCircularPointRelCentre(middleRadius, centreAngle);
  drawTriangle(ctx, p1, p2, p3, strokeColour, lineWidth, filled);
}

// /**
//  * Draws a rectangle
//  * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
//  * @param {Vector} topLeftPoint - the top left point of the rectangle
//  * @param {number} width - the width of the rectangle
//  * @param {number} height - the height of the rectangle
//  * @param {string} strokeColour - the colour of the rectangle
//  * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
//  */
// function drawRectangle(ctx, topLeftPoint, width, height, strokeColour, filled) {
//   ctx.beginPath();
//   ctx.rect(topLeftPoint.x(), topLeftPoint.y(), width, height);
//   applyStyleAndStroke(ctx, strokeColour, 0, filled);
// }

// /**
//  * Draws a square
//  * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
//  * @param {Vector} topLeftPoint - the top left point of the square
//  * @param {number} width - the width of the square
//  * @param {string} strokeColour - the colour of the square
//  * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
//  */
// function drawSquare(ctx, topLeftPoint, width, strokeColour, filled) {
//   drawRectangle(ctx, topLeftPoint, width, width, strokeColour, filled);
// }

function drawText(ctx, text, position, fontSize, colour, centreFont) {
  ctx.font = fontSize + "px Advantage";

  if (centreFont) {
    ctx.textAlign = "center";
    // Adjust for font size
    const FONT_OFFSET = ctx.measureText("0").width / 2;
    position.setY(position.y() + FONT_OFFSET);
  }

  ctx.fillStyle = colour;
  ctx.fillText(text, position.x(), position.y());
}

/**
 * Applies styling to a drawn-on canvas context and renders it
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function applyStyleAndStroke(ctx, strokeColour, lineWidth, filled) {
  if (filled) {
    ctx.fillStyle = strokeColour;
    ctx.fill();
  }

  ctx.strokeStyle = strokeColour;
  ctx.lineWidth = lineWidth;
  ctx.stroke();
}

/////////////////////
// Maths Functions //
/////////////////////

/**
 * Returns the coordinates of a point on a circle
 * @param {number} radius - the radius of the circle
 * @param {number} angle - the angle along the circle the point is
 * @returns {Vector} the point
 */
function getCircularPoint(radius, angle) {
  return new Vector(radius*Math.cos(angle), radius*Math.sin(angle));
}

/**
 * Returns the coordinates of a point on a circle, relative to the centre point of the UI.
 * In essence, it calculates the circular point's coordinates and translates it by the centre point
 * @param radius
 * @param angle
 * @returns {Vector}
 */
function getCircularPointRelCentre(radius, angle) {
  return Vector.add(CENTRE, getCircularPoint(radius, angle));
}