/////////////
// Classes //
/////////////

/**
 * A class representing a 2D vector
 */
class Vector {
  constructor(x, y) {
    this.array = [x, y];
  }

  /**
   * Reverse the signs of all elements in the vector
   */
  negate() {
    this.array.forEach(e => -e);
  }

  static mapOperations(vector, amount, func) {
    let result = new Vector(vector.array[0], vector.array[1]);

    if (amount instanceof Vector) {
      for (let i = 0; i < result.array.length; i++)
        result.array[i] = func(result.array[i], amount.array[i]);
    } else {
      result.array = result.array.map((e) => func(e, amount));
    }

    return result;
  }

  /**
   * Adds a vector from another or a scalar value
   * @param {Vector} vector - the vector to be added to
   * @param {Vector|number} amount - the vector/scalar to add
   * @returns {Vector} - the resulting vector
   */
  static add(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e+a);
  }

  /**
   * Subtracts a vector from another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static sub(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e-a);
  }

  /**
   * Multiples a vector with another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static mult(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e*a);
  }

  /**
   * Divides a vector by another or a scalar value
   * @param {Vector} vector - the vector to be divided by
   * @param {Vector|number} amount - the vector/scalar to divide with
   * @returns {Vector} - the resulting vector
   */
  static div(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e/a);
  }

  /**
   * Adds to this vector another vector from another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to add
   */
  add(amount) {
    this.array = Vector.add(this, amount).array;
  }

  /**
   * Subtracts this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  sub(amount) {
    this.array = Vector.sub(this, amount).array;
  }

  /**
   * Multiples this vector with another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  mult(amount) {
    this.array = Vector.mult(this, amount).array;
  }

  /**
   * Divides this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to divide with
   */
  div(amount) {
    this.array = Vector.div(this, amount).array;
  }

  /**
   * Normalises this vector
   */
  normalise() {
    let mag = this.mag();

    if (mag > 0)
      this.div(mag);
  }

  /**
   * Calculates and returns this vector's magnitude
   * @returns {number} - the magnitude of this vector
   */
  mag() {
    return Math.sqrt(this.array.reduce((a, b) => Math.pow(a, 2) + Math.pow(b, 2)));
  }

  /**
   * Sets this vector's components to that of another's
   * @param {Vector} vector - the components to copy
   */
  set(vector) {
    for (let i = 0; i < this.array.length; i++)
      this.array[i] = vector.array[i];
  }

  x() { return this.array[0]; }
  y() { return this.array[1]; }
}


///////////////
// Constants //
///////////////
const CENTRE = new Vector(220, 220);

// The width of stroke effects for various components
const WIDTH = {
  RPM: 70,
  BG_STROKE: 5
}

// The radius values for various components
const RADII = {base: 220};
RADII.bgOuter = RADII.base - WIDTH.BG_STROKE;
RADII.bgInner = RADII.bgOuter - WIDTH.RPM;
RADII.rpm = RADII.bgOuter - WIDTH.RPM/2;
RADII.rpmInner = 100;

// The colour scheme
const ACCENT_COLOUR = "rgb(255, 255, 255)";
const BASE_COLOUR = "rgb(0, 180, 255)";

// The starting and end angle of the RPM counter
const RPM_START_ANGLE = 0.75*Math.PI;
const RPM_END_ANGLE = 2.25*Math.PI;

const INNER_TEXT_OFFSET = new Vector(0, 35);


//////////////////////////////
// Rendering/Main Functions //
//////////////////////////////

function draw() {
  let canvas = document.getElementById("canvas");
  // Ensure correct HD resolution on the canvas (fix for blurry text)
  canvas.width = canvas.getBoundingClientRect().width;
  canvas.height = canvas.getBoundingClientRect().height;
  let ctx = canvas.getContext("2d");


  drawBackground(ctx);
  drawRPM(ctx, 4000, 6000);
  drawInterior(ctx, 100, "2", "3", "4");
}

function drawBackground(ctx) {
  // Draw outer border
  drawArc(ctx, CENTRE, RADII.bgOuter, RPM_START_ANGLE, RPM_END_ANGLE, ACCENT_COLOUR, WIDTH.BG_STROKE);
  // Draw inner border
  // drawArc(ctx, CENTRE, RADII.bgInner, RPM_START_ANGLE, RPM_END_ANGLE, ACCENT_COLOUR, WIDTH.BG_STROKE);


  // Draw the end lines and connect them together in the centre
  let endPoint1 = Vector.add(getCircularPoint(RADII.rpmInner, RPM_START_ANGLE), CENTRE);
  let startPoint1 = getStartRPMLine(RPM_START_ANGLE, endPoint1);
  let endPoint2 = Vector.add(getCircularPoint(RADII.rpmInner,   RPM_END_ANGLE), CENTRE);
  let startPoint2 = getStartRPMLine(RPM_END_ANGLE, endPoint2);

  drawLines(ctx, [startPoint1, endPoint1, endPoint2, startPoint2], ACCENT_COLOUR, WIDTH.BG_STROKE);
}

function drawInterior(ctx, speed, prevGear, currGear, nextGear) {
  drawSpeed(ctx, speed);
  drawGear(ctx, prevGear, currGear, nextGear);

  const INTERIOR_CENTRE = Vector.sub(CENTRE, new Vector(0, 35/2 + WIDTH.BG_STROKE));

  // Draw dividing line
  const DIV_RADIUS = new Vector(100, 0);
  const DIV_WIDTH = 3;
  drawLine(
      ctx,
      Vector.sub(INTERIOR_CENTRE, DIV_RADIUS),
      Vector.add(INTERIOR_CENTRE, DIV_RADIUS),
      ACCENT_COLOUR,
      DIV_WIDTH
  );
}

function drawSpeed(ctx, speed) {
  const SPEED_POS = Vector.sub(CENTRE, INNER_TEXT_OFFSET);

  ctx.font = "80px Digital";
  ctx.textAlign = "center";
  // Draw outline
  ctx.fillStyle = "rgba(255, 255, 255, 0.25)";
  ctx.fillText("000", SPEED_POS.x(), SPEED_POS.y());
  // Draw actual speed
  ctx.fillStyle = "rgba(255, 255, 255)";
  ctx.fillText(speed, SPEED_POS.x(), SPEED_POS.y());
}

function drawGear(ctx, prevGear, currGear, nextGear) {
  const GEAR_POS = Vector.add(CENTRE, INNER_TEXT_OFFSET);

  const ADJ_GEAR_OPACITY = "rgba(255, 255, 255, 0.25)";

  // A gradient that has max opacity for the middle gear, and lower opacity for previous and next gears
  let gradient = ctx.createLinearGradient(GEAR_POS.x() - 40, GEAR_POS.y(), GEAR_POS.x() + 40, GEAR_POS.y());
  gradient.addColorStop(0, ADJ_GEAR_OPACITY);
  gradient.addColorStop(0.33, ADJ_GEAR_OPACITY);
  gradient.addColorStop(0.33, "rgba(255, 255, 255, 1)");
  gradient.addColorStop(0.66, "rgba(255, 255, 255, 1)");
  gradient.addColorStop(0.66, ADJ_GEAR_OPACITY);
  gradient.addColorStop(1, ADJ_GEAR_OPACITY);

  // Draw
  ctx.font = "55px Digital";
  ctx.textAlign = "center";
  ctx.fillStyle = gradient;
  ctx.fillText("" + prevGear + currGear + nextGear, GEAR_POS.x(), GEAR_POS.y());
}

/**
 * Returns the starting point for an end on the RPM dial
 * @param {number} angle - the angle of the end of the RPM dial to close off
 * @param {Vector} endPoint - the end point of the line to draw
 * @return {Vector} the starting point
 */
function getStartRPMLine(angle, endPoint) {
  return Vector.add(getCircularPoint(RADII.base - WIDTH.BG_STROKE/2, angle), CENTRE);
}

function drawRPM(ctx, rpm, rpmMax) {
  // The angle at which maximum, non-redline RPM occurs
  const NON_REDLINE_MAX_ANGLE = 2*Math.PI;

  let dif = RPM_END_ANGLE - RPM_START_ANGLE;
  let ratio = Math.min(1, rpm / rpmMax);
  // The angle the RPM value reaches
  let endAngle = RPM_START_ANGLE + ratio*dif;

  const ARC_WIDTH = WIDTH.RPM - WIDTH.BG_STROKE;

  // The background arc showing the RPM area
  drawArc(ctx, CENTRE, RADII.RPM_BG, RPM_START_ANGLE, RPM_END_ANGLE, "rgb(0,0,0)", ARC_WIDTH);



  // The red-line RPM
  drawArc(ctx, CENTRE, RADII.RPM_BG, NON_REDLINE_MAX_ANGLE, RPM_END_ANGLE, "rgb(255,0,0)", ARC_WIDTH);
  // The arc to represent the current RPM
  drawArc(ctx, CENTRE, RADII.RPM_BG, RPM_START_ANGLE, endAngle, BASE_COLOUR, WIDTH.RPM/2);
}


///////////////////////
// Drawing Functions //
///////////////////////

/**
 * Draws a number of lines along the points specified, in that order
 * @param ctx
 * @param {Vector[]} points - the points the lines will be draw through
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function drawLines(ctx, points, strokeColour, lineWidth) {
  ctx.beginPath();

  let p1 = points[0];
  ctx.moveTo(p1.x(), p1.y());

  for (let i = 1; i < points.length; i++) {
    let point = points[i];
    ctx.lineTo(point.x(), point.y());
  }

  applyStyleAndStroke(ctx, strokeColour, lineWidth);
}

/**
 * Draws a line that connects two points
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} point1 - the first point
 * @param {Vector} point2 - the second point
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function drawLine(ctx, point1, point2, strokeColour, lineWidth) {
  ctx.beginPath();
  ctx.moveTo(point1.x(), point1.y());
  ctx.lineTo(point2.x(), point2.y());
  applyStyleAndStroke(ctx, strokeColour, lineWidth);
}

/**
 * Draws an arc between two angles
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the arc
 * @param {number} radius - the radius of the arc
 * @param {number} startAngle - the starting angle of the arc
 * @param {number} endAngle -  the ending angle of the arc
 * @param {string} strokeColour - the colour of the arc
 * @param {number} lineWidth - the width of each arc
 */
function drawArc(ctx, centre, radius, startAngle, endAngle, strokeColour, lineWidth) {
  ctx.beginPath();
  ctx.arc(centre.x(), centre.y(), radius, startAngle, endAngle);
  applyStyleAndStroke(ctx, strokeColour, lineWidth);
}

/**
 * Applies styling to a drawn-on canvas context and renders it
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function applyStyleAndStroke(ctx, strokeColour, lineWidth) {
  ctx.strokeStyle = strokeColour;
  ctx.lineWidth = lineWidth;
  ctx.stroke();
}

/////////////////////
// Maths Functions //
/////////////////////

/**
 * Returns the coordinates of a point on a circle
 * @param {number} radius - the radius of the circle
 * @param {number} angle - the angle along the circle the point is
 * @returns {Vector} the point
 */
function getCircularPoint(radius, angle) {
  return new Vector(radius*Math.cos(angle), radius*Math.sin(angle));
}