/////////////
// Classes //
/////////////

/**
 * A class representing a 2D vector
 */
class Vector {
  constructor(x, y) {
    this.array = [x, y];
  }

  /**
   * Reverse the signs of all elements in the vector
   */
  negate() {
    this.array.forEach(e => -e);
  }

  static mapOperations(vector, amount, func) {
    let result = new Vector(vector.array[0], vector.array[1]);

    if (amount instanceof Vector) {
      for (let i = 0; i < result.array.length; i++)
        result.array[i] = func(result.array[i], amount.array[i]);
    } else {
      result.array = result.array.map((e) => func(e, amount));
    }

    return result;
  }

  /**
   * Adds a vector from another or a scalar value
   * @param {Vector} vector - the vector to be added to
   * @param {Vector|number} amount - the vector/scalar to add
   * @returns {Vector} - the resulting vector
   */
  static add(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e+a);
  }

  /**
   * Subtracts a vector from another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static sub(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e-a);
  }

  /**
   * Multiples a vector with another or a scalar value
   * @param {Vector} vector - the vector to be subtracted from
   * @param {Vector|number} amount - the vector/scalar to subtract
   * @returns {Vector} - the resulting vector
   */
  static mult(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e*a);
  }

  /**
   * Divides a vector by another or a scalar value
   * @param {Vector} vector - the vector to be divided by
   * @param {Vector|number} amount - the vector/scalar to divide with
   * @returns {Vector} - the resulting vector
   */
  static div(vector, amount) {
    return Vector.mapOperations(vector, amount, (e, a) => e/a);
  }

  /**
   * Adds to this vector another vector from another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to add
   */
  add(amount) {
    this.array = Vector.add(this, amount).array;
  }

  /**
   * Subtracts this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  sub(amount) {
    this.array = Vector.sub(this, amount).array;
  }

  /**
   * Multiples this vector with another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to subtract
   */
  mult(amount) {
    this.array = Vector.mult(this, amount).array;
  }

  /**
   * Divides this vector by another or a scalar value
   * @param {Vector|number} amount - the vector/scalar to divide with
   */
  div(amount) {
    this.array = Vector.div(this, amount).array;
  }

  /**
   * Normalises this vector
   */
  normalise() {
    let mag = this.mag();

    if (mag > 0)
      this.div(mag);
  }

  /**
   * Calculates and returns this vector's magnitude
   * @returns {number} - the magnitude of this vector
   */
  mag() {
    return Math.sqrt(this.array.reduce((a, b) => Math.pow(a, 2) + Math.pow(b, 2)));
  }

  /**
   * Sets this vector's components to that of another's
   * @param {Vector} vector - the components to copy
   */
  set(vector) {
    for (let i = 0; i < this.array.length; i++)
      this.array[i] = vector.array[i];
  }

  x() { return this.array[0]; }
  y() { return this.array[1]; }
}


///////////////
// Constants //
///////////////

// The width of stroke effects for various components
const WIDTH = {
  // The current RPM arc that goes around the outside
  RPM_ARC: 30,
  // The interior's stroke between itself and the RPM background
  INNER_STROKE: 5,
  // The width of each tick on the RPM gauge
  RPM_TICK_WIDTH: 8
}

// The background to the RPM thousand numbers
WIDTH.RPM_BG = 80 - WIDTH.RPM_ARC;

// The centre radius values for various components
// Any stroke effect adds thickness to a radius
const RADII = {
  BASE: 150,
  BG_INNER: 90,
  RPM_DIAL_RADIUS: 20
};
RADII.RPM_ARC_BG = RADII.BG_INNER + WIDTH.RPM_BG/2;
RADII.RPM_ARC = RADII.BASE - WIDTH.RPM_ARC/2;
RADII.RPM_INNER = RADII.BASE - WIDTH.RPM_ARC;

// The size of the box the gear number is in
WIDTH.GEAR_BOX = RADII.BG_INNER - WIDTH.INNER_STROKE;


const CENTRE = new Vector(150, 150);

// The colour scheme
const ACCENT_COLOUR = "rgb(255, 128, 0)";
const RPM_COLOUR    = "rgb(255, 255, 0)";
const BASE_COLOUR   = "rgb(0, 0, 0)";

// The starting and end angle of the RPM counter
const RPM_START_ANGLE = 0.5*Math.PI;
const RPM_END_ANGLE = 2*Math.PI;
// The angle at which maximum, non-redline RPM occurs
const NON_REDLINE_MAX_ANGLE = 1.75*Math.PI;


//////////////////////////////
// Rendering/Main Functions //
//////////////////////////////

let rpm = 0;
let dir = 1;
let speed = 0;

function draw() {
  let canvas = document.getElementById("canvas");
  // Ensure correct HD resolution on the canvas (fix for blurry text)
  canvas.width = canvas.getBoundingClientRect().width;
  canvas.height = canvas.getBoundingClientRect().height;
  let ctx = canvas.getContext("2d");

  drawInterior(ctx, "2", 123);
  drawRPM(ctx, 500, 13000);


  // let interval = 1000/60;
  // setInterval(function() {
  //   rpm += interval * 1.1 * dir;
  //   speed += Math.floor(interval/10) * dir;
  //
  //   if (rpm >= 8350-interval || rpm <= interval)
  //     dir = -dir;
  //
  //   ctx.clearRect(0, 0, canvas.width, canvas.height);
  //   drawInterior(ctx, "2", speed);
  //   drawRPM(ctx, rpm, 8450);
  // }, interval);
}

function drawInterior(ctx, currGear, speed) {
  const STROKE_WIDTH = 10;
  const RADIUS = RADII.BG_INNER - STROKE_WIDTH/2;

  // Draw interior arc between RPM background and the interior
  drawArc(ctx, CENTRE, RADIUS, RPM_START_ANGLE, RPM_END_ANGLE, BASE_COLOUR, STROKE_WIDTH, false);

  drawGear(ctx, currGear);
  drawSpeed(ctx, speed);
}

function drawGear(ctx, currGear) {
  // Draw the box that the gear will sit in
  drawWedge(ctx, CENTRE, WIDTH.GEAR_BOX, RPM_END_ANGLE, RPM_START_ANGLE, BASE_COLOUR, 0, true);

  // The angle at which a line goes through the diagonal of the gear's box
  const HALF_ANGLE = RPM_END_ANGLE + RPM_START_ANGLE/2;


  const P1 = getCircularPointRelCentre(WIDTH.GEAR_BOX, HALF_ANGLE);
  const P2 = getCircularPointRelCentre(RADII.RPM_DIAL_RADIUS, HALF_ANGLE);
  // The point at the centre of the box (accounting for RPM pointer and wedge gear box shape)
  let boxCentre = Vector.sub(P1, P2);
  boxCentre.div(2);
  boxCentre.add(P2);

  ctx.font = "60px Advantage";
  ctx.textAlign = "center";
  // Adjust for font size
  const FONT_OFFSET = ctx.measureText("0").width/2;
  boxCentre.add(new Vector(0, FONT_OFFSET));

  ctx.fillStyle = RPM_COLOUR;
  ctx.fillText(currGear, boxCentre.x(), boxCentre.y());
}

function drawSpeed(ctx, speed) {
  const FONT_SIZE = 55;
  const SPEED_POS = Vector.sub(CENTRE, new Vector(0, FONT_SIZE/2));

  ctx.font = FONT_SIZE+"px Advantage";
  ctx.textAlign = "center";
  ctx.fillStyle = RPM_COLOUR;
  ctx.fillText(speed, SPEED_POS.x(), SPEED_POS.y());
}

function drawRPM(ctx, currRPM, maxRPM) {
  let dif = RPM_END_ANGLE - RPM_START_ANGLE;
  let ratio = Math.min(1, currRPM / maxRPM);
  // The angle the RPM value reaches
  let currRPMAngle = RPM_START_ANGLE + ratio*dif;

  // The background arc showing the RPM areas
  drawArc(ctx, CENTRE, RADII.RPM_ARC_BG, RPM_START_ANGLE, RPM_END_ANGLE, "rgb(32,32,32)", WIDTH.RPM_BG);
  // The current RPM and background of the RPM
  drawRPMStroke(ctx, currRPMAngle);
  // The ticks and outer-most outline
  drawRPMTicksAndOutline(ctx, maxRPM);
  // The stroke on the outside of the RPM gauge
  drawArc(ctx, CENTRE, RADII.BASE-2, RPM_START_ANGLE, RPM_END_ANGLE, ACCENT_COLOUR, 4);
  // The pointer in the centre
  drawRPMPointer(ctx, currRPMAngle);
}

function drawRPMPointer(ctx, angleRPM) {
  // The width of the pointer at its maximum point
  const ANGLE_WIDTH = 0.125*Math.PI;

  // The angle at which the current RPM equals on the opposite side
  let angleRPMInv = angleRPM + Math.PI;

  // Draw triangular piece opposite to the RPM pointed at
  drawTriangleOnCircularBase(ctx, angleRPMInv, ANGLE_WIDTH, RADII.RPM_DIAL_RADIUS, RADII.RPM_DIAL_RADIUS*2, ACCENT_COLOUR, 1, true);
  // Draw triangular piece that points to the current RPM value
  drawTriangleOnCircularBase(ctx, angleRPM, ANGLE_WIDTH, RADII.RPM_DIAL_RADIUS, RADII.BASE, ACCENT_COLOUR, 1, true);

  // Draw centre circle part
  drawCircle(ctx, CENTRE, RADII.RPM_DIAL_RADIUS, ACCENT_COLOUR, 1, true);
  drawCircle(ctx, CENTRE, RADII.RPM_DIAL_RADIUS/2, "black", 1, true);

  // Draw red accents onto the pointer
  const ANGLE_WIDTH_ACCENTS = ANGLE_WIDTH/4;
  drawTriangleOnCircularBase(ctx, angleRPM, ANGLE_WIDTH_ACCENTS, RADII.RPM_DIAL_RADIUS, RADII.BASE, "red", 1, true);
  drawTriangleOnCircularBase(ctx, angleRPMInv, ANGLE_WIDTH_ACCENTS, RADII.RPM_DIAL_RADIUS, RADII.RPM_DIAL_RADIUS*2, "red", 1, true);
}

function drawRPMStroke(ctx, angleRPM) {
  // Draw non-redline background RPM
  drawArc(ctx, CENTRE, RADII.RPM_ARC, RPM_START_ANGLE, NON_REDLINE_MAX_ANGLE, BASE_COLOUR, WIDTH.RPM_ARC);
  // The current RPM
  drawArc(ctx, CENTRE, RADII.RPM_ARC, RPM_START_ANGLE, angleRPM, RPM_COLOUR, WIDTH.RPM_ARC);
  // Draw the red-line RPM (will overwrite the current RPM if it goes into red-line)
  drawArc(ctx, CENTRE, RADII.RPM_ARC, NON_REDLINE_MAX_ANGLE, RPM_END_ANGLE, "red", WIDTH.RPM_ARC);
}

function drawRPMTicksAndOutline(ctx, maxRPM) {
  // The RPM spacing between two ticks
  const TICK_SPACING = 250;
  const NUM_TICKS = Math.floor(maxRPM / TICK_SPACING);
  // The angle spacing between two ticks
  const SPACING = (RPM_END_ANGLE - RPM_START_ANGLE) / NUM_TICKS;

  let angle = RPM_START_ANGLE;

  for (let i = 0; i <= NUM_TICKS; i++) {
    let rpmOfCurrAngle = i * TICK_SPACING;
    let isAThousand = rpmOfCurrAngle % 1000 === 0;

    if (isAThousand) {
      let numThousand = (TICK_SPACING / 1000) * i;
      labelRPMThousand(ctx, numThousand, angle, i === 0 || rpmOfCurrAngle === maxRPM);
      addRPMTick(ctx, angle, isAThousand);
    }

    angle += SPACING;
  }
}

function getRadiusOfRPMAt(isInner, isPadded) {
  const PADDING = WIDTH.RPM_TICK_WIDTH/2;

  if (isInner) {
    let radius = RADII.RPM_INNER;
    return isPadded ? radius+PADDING : radius;
  } else {
    let radius = RADII.BASE;
    return isPadded ? radius-PADDING : radius;
  }
}

function labelRPMThousand(ctx, num, angle, isEndingRPM) {
  const FONT_SIZE = 25;

  ctx.font = FONT_SIZE+"px Advantage";
  ctx.textAlign = "center";
  ctx.fillStyle = angle > NON_REDLINE_MAX_ANGLE ? "red" : ACCENT_COLOUR;
  const FONT_OFFSET = new Vector(0, ctx.measureText("0").width/2);

  let coords = getCircularPointRelCentre(RADII.RPM_ARC_BG - FONT_SIZE/2, angle);
  coords.add(FONT_OFFSET);
  ctx.fillText(num, coords.x(), coords.y());

  addRPMTick(ctx, angle, !isEndingRPM);
}

function addRPMTick(ctx, angle, isOuterPadded) {
  let p1 = getCircularPointRelCentre(getRadiusOfRPMAt(false, isOuterPadded), angle);
  let p2 = getCircularPointRelCentre(getRadiusOfRPMAt(true, false), angle);
  drawLine(ctx, p1, p2, ACCENT_COLOUR, WIDTH.RPM_TICK_WIDTH);
}


///////////////////////
// Drawing Functions //
///////////////////////

/**
 * Draws a number of lines along the points specified, in that order
 * @param ctx
 * @param {Vector[]} points - the points the lines will be draw through
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function drawLines(ctx, points, strokeColour, lineWidth) {
  ctx.beginPath();

  let p1 = points[0];
  ctx.moveTo(p1.x(), p1.y());

  for (let i = 1; i < points.length; i++) {
    let point = points[i];
    ctx.lineTo(point.x(), point.y());
  }

  applyStyleAndStroke(ctx, strokeColour, lineWidth, false);
}

function drawArcs(ctx, arcCentre, arcs, strokeColour, lineWidth) {
  ctx.beginPath();

  for (let i = 1; i < arcs.length; i++) {
    let arc = arcs[i];
    ctx.arc(arcCentre.x(), arcCentre.y(), arc.radius, arc.startAngle, arc.endAngle);
  }

  applyStyleAndStroke(ctx, strokeColour, lineWidth, false);
}

/**
 * Draws a line that connects two points
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} point1 - the first point
 * @param {Vector} point2 - the second point
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 */
function drawLine(ctx, point1, point2, strokeColour, lineWidth) {
  ctx.beginPath();
  ctx.moveTo(point1.x(), point1.y());
  ctx.lineTo(point2.x(), point2.y());
  applyStyleAndStroke(ctx, strokeColour, lineWidth, false);
}

/**
 * Draws an arc between two angles
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the arc
 * @param {number} radius - the radius of the arc
 * @param {number} startAngle - the starting angle of the arc
 * @param {number} endAngle -  the ending angle of the arc
 * @param {string} strokeColour - the colour of the arc
 * @param {number} lineWidth - the width of the arc
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawArc(ctx, centre, radius, startAngle, endAngle, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.arc(centre.x(), centre.y(), radius, startAngle, endAngle);
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a circle at a specific point
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the circle
 * @param {number} radius - the radius of the circle
 * @param {string} strokeColour - the colour of the circle
 * @param {number} lineWidth - the thickness of the line of the circle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawCircle(ctx, centre, radius, strokeColour, lineWidth, filled) {
  drawArc(ctx, centre, radius, 0, 2*Math.PI, strokeColour, lineWidth, filled);
}

/**
 * Draws an arc between two angles and fills it in, relative to its centre origin, to create a
 * wedge shape
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} centre - the centre point of the wedge
 * @param {number} radius - the radius of the wedge
 * @param {number} startAngle - the starting angle of the wedge
 * @param {number} endAngle -  the ending angle of the wedge
 * @param {string} strokeColour - the colour of the wedge
 * @param {number} lineWidth - the width of the wedge
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawWedge(ctx, centre, radius, startAngle, endAngle, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.moveTo(CENTRE.x(), CENTRE.y());
  ctx.arc(centre.x(), centre.y(), radius, startAngle, endAngle);
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a triangle, specified by its three vertices
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} point1 - the first vertex
 * @param {Vector} point2 - the second vertex
 * @param {Vector} point3- the second vertex
 * @param {string} strokeColour - the colour of the triangle
 * @param {number} lineWidth - the thickness of the line of the triangle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawTriangle(ctx, point1, point2, point3, strokeColour, lineWidth, filled) {
  ctx.beginPath();
  ctx.moveTo(point1.x(), point1.y());
  ctx.lineTo(point2.x(), point2.y());
  ctx.lineTo(point3.x(), point3.y());
  applyStyleAndStroke(ctx, strokeColour, lineWidth, filled);
}

/**
 * Draws a triangle whose base two vertices are positioned on a circle, and whose third vertex is
 * extended on another circle
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {number} centreAngle - the angle of the circle at the centre point of the triangle
 * @param {number} angleDev - the amount past the centre angle the triangle will extend to
 * @param {number} circleRadius - the circle's radius
 * @param {number} middleRadius - the radius the top-most vertex of the triangle will be placed on
 * @param {string} strokeColour - the colour of the triangle
 * @param {number} lineWidth - the thickness of the line of the triangle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawTriangleOnCircularBase(ctx, centreAngle, angleDev, circleRadius, middleRadius, strokeColour, lineWidth, filled) {
  let p1 = Vector.add(CENTRE, getCircularPoint(circleRadius, centreAngle - angleDev));
  let p2 = Vector.add(CENTRE, getCircularPoint(circleRadius, centreAngle + angleDev));
  let p3 = Vector.add(CENTRE, getCircularPoint(middleRadius, centreAngle));
  drawTriangle(ctx, p1, p2, p3, strokeColour, lineWidth, filled);
}

/**
 * Draws a rectangle
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} topLeftPoint - the top left point of the rectangle
 * @param {number} width - the width of the rectangle
 * @param {number} height - the height of the rectangle
 * @param {string} strokeColour - the colour of the rectangle
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawRectangle(ctx, topLeftPoint, width, height, strokeColour, filled) {
  ctx.beginPath();
  ctx.rect(topLeftPoint.x(), topLeftPoint.y(), width, height);
  applyStyleAndStroke(ctx, strokeColour, 0, filled);
}

/**
 * Draws a square
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {Vector} topLeftPoint - the top left point of the square
 * @param {number} width - the width of the square
 * @param {string} strokeColour - the colour of the square
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function drawSquare(ctx, topLeftPoint, width, strokeColour, filled) {
  drawRectangle(ctx, topLeftPoint, width, width, strokeColour, filled);
}

/**
 * Applies styling to a drawn-on canvas context and renders it
 * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
 * @param {string} strokeColour - the colour of the lines
 * @param {number} lineWidth - the width of each line
 * @param {boolean} filled - whether the shape is filled in (strokeColour will be used)
 */
function applyStyleAndStroke(ctx, strokeColour, lineWidth, filled) {
  if (filled) {
    ctx.fillStyle = strokeColour;
    ctx.fill();
  }

  ctx.strokeStyle = strokeColour;
  ctx.lineWidth = lineWidth;
  ctx.stroke();
}

/////////////////////
// Maths Functions //
/////////////////////

/**
 * Returns the coordinates of a point on a circle
 * @param {number} radius - the radius of the circle
 * @param {number} angle - the angle along the circle the point is
 * @returns {Vector} the point
 */
function getCircularPoint(radius, angle) {
  return new Vector(radius*Math.cos(angle), radius*Math.sin(angle));
}

/**
 * Returns the coordinates of a point on a circle, relative to the centre point of the UI.
 * In essence, it calculates the circular point's coordinates and translates it by the centre point
 * @param radius
 * @param angle
 * @returns {Vector}
 */
function getCircularPointRelCentre(radius, angle) {
  return Vector.add(CENTRE, getCircularPoint(radius, angle));
}