angular.module("beamng.apps")
/**
 * The base service for the future variant apps
 */
.factory("ModVelocityBase", ["TachoBase", function(TachoBase) {
  /**
   *
   */
  class VelocityGaugeBuilder extends TachoBase.GaugeBuilder {
    /**
     * Constructor
     * @param {number} baseRadius - the radius of the gauge at its maximum point
     * @param {ColourScheme} colourScheme - the colour scheme to apply to the gauge
     * @param {Range} angleRange - the range of angles of the dial's arc overall
     */
    constructor(baseRadius, colourScheme, angleRange) {
      super();
      this.baseRadius = baseRadius;
      this.colourScheme = colourScheme;
      this.angleRange = angleRange;
    }

    /**
     *
     * @param dialRadius
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createPointer(dialRadius) {
      // The width of the pointer at its maximum point
      const MAJ_WIDTH = 0.1*Math.PI;
      // The radius of the smaller pointer on the opposite side
      const INV_POINTER_RADIUS = dialRadius*1.5;
      // Draw accents onto the pointer
      const MIN_WIDTH = MAJ_WIDTH/4;

      let majPointRadii = new TachoBase.Range(dialRadius, this.baseRadius);
      let minPointRadii = new TachoBase.Range(dialRadius, INV_POINTER_RADIUS);
      let accent = this.colourScheme.ACCENT;

      return this.setAndcreatePointer(majPointRadii, minPointRadii, MAJ_WIDTH, MIN_WIDTH, accent, dialRadius);
    }

    /**
     * Creates the GaugeTicks instance for the gauge
     * @param {number} numTicks - the number of ticks to generate
     * @param {number} tickSpacing - the spacing between two ticks, in terms of value
     * @param {number} arcWidth - the width of the arc
     * @param {number} minTickVal - the lowest tick's value. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createTicks(numTicks, tickSpacing, arcWidth, minTickVal) {
      return this.setTicks(new TachoBase.GaugeTicks(
          numTicks,
          tickSpacing,
          2,
          4,
          this.colourScheme.ACCENT,
          // +-2 to not go over bounds
          new TachoBase.Range(this.baseRadius - arcWidth + 2, this.baseRadius - 2),
          this.angleRange,
          false,
          minTickVal
      ));
    }

    /**
     * Abstract
     * Must be implemented by a base class
     * Is allowed to have different parameters/number of parameters
     */
    createArcs() {
      console.error("createArcs() must be implemented in base class!");
    }

    /**
     * Creates the GaugeNumbers instance for the gauge
     * @param {number} numNumbers - the number of numbers that will fill the major points on the
     * tick gauge
     * @param {number} fontSize - the size of the font for the numbers
     * @param {number} numbersRadius - the radius of the arc that the numbers will be spaced around
     * @param {number} spacing - the spacing between two numbers
     * @param {number} [startNum] - the starting number. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createNumbers(numNumbers, fontSize, numbersRadius, spacing, startNum) {
      return this.setNumbers(new TachoBase.GaugeNumbers(
          numNumbers,
          spacing,
          this.angleRange,
          new TachoBase.Font("Speed12", fontSize, true),
          numbersRadius,
          this.colourScheme.TEXT.secondary,
          startNum
      ));
    }
  }


  /**
   * Returns the the angle range for the middle-right rounded rectangle {@see drawRoundRectRight()}
   * 's arc
   * @param {BaseApp} app - the app intending to draw the round rectangle polygon using this angle
   * @returns {Range}
   */
  this.roundRectRightAngle = function(app) {
    return new TachoBase.Range(
        app.gauge.angleRange.lower * 0.125, app.gauge.angleRange.lower * 0.425
    );
  }

  /**
   * Draws (but does not complete) a rounded rectangle at the mid-right position of the UI
   * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
   * @param {BaseApp} app - the app drawing this polygon
   * @param {number} leftX - the left-most x coordinate
   */
  this.drawRoundRectRight = function(ctx, app, leftX) {
    ctx.beginPath();
    new TachoBase.RoundCornerRectangleBR(app.radius, this.roundRectRightAngle(app), app.CENTRE, leftX)
        .drawContinue(ctx);
    ctx.lineWidth = 1;
    ctx.fillStyle = "black";
  }


  this.TachoBase = TachoBase;
  this.VelocityGaugeBuilder = VelocityGaugeBuilder;


  return this;
}]);